// Copyright (C) 1997, 1998 by DI Wolfgang Platzer, IAIK
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//

package org.w3c.jigsaw.http.socket;

import org.w3c.jigsaw.http.*;
import org.w3c.jigsaw.config.PropertySet;

import java.net.ServerSocket;
import java.io.*;
import java.util.Vector;
import java.security.*;
import java.security.cert.*;
import java.math.BigInteger;

import javax.crypto.spec.*;

import iaik.jigsaw.https.*;

import iaik.utils.KeyAndCertificate;
import iaik.security.ssl.*;
import iaik.security.rsa.RSAPrivateKey;
import iaik.asn1.structures.Name;
import iaik.pkcs.pkcs8.PrivateKeyInfo;
import iaik.x509.X509Certificate;

/**
 * This class extends SocketClientFactory and adds the SSL functionality.
 *
 * @see httpsd
 * @see SSLSocketClient
 *
 * @autor wplatzer
 */
public class SSLSocketClientFactory extends SocketClientFactory {

  protected SSLServerContext ssl_context;
  protected JigsawServerTrustDecider trust_decider;
  protected httpsd ssl_server;

  // parameters for client authentication
  private boolean client_authentication = false;
  private boolean require_certificate = false;
  private boolean require_trusted_root = false;
  private String trusted_signers_file_name;

  /**
   * If this property is set to "true" the client <b>may</b> send a certificate.
   */
  public static final String 
  CLIENT_AUTHENTICATION_P = "iaik.jigsaw.ssl.client.authentication";
  
  /**
   * If this property is set to "true" the client <b>must</b> send a certificate.
   */
  public static final String 
  CLIENT_REQUIRE_CERTIFICATE_P = "iaik.jigsaw.ssl.client.requireCertificate";

  /**
   * If this property is set to "true" the client <b>must</b> send a certificate,
   * which was issued from a trusted root.<p>
   * The trusted root certificates can be specified with TRUSTET_SIGNERS_P.
   *
   * @see TRUSTET_SIGNERS_P
   */
  public static final String 
  CLIENT_REQUIRE_TRUSTED_ROOT_P = "iaik.jigsaw.ssl.client.requireTrustedRoot";

  /**
   * Name of a file which contains the trusted certificates which signed
   * the user certificates for client authentication.
   */
  public static final String 
  TRUSTET_SIGNERS_P = "iaik.jigsaw.ssl.trustedSigners";
  
  /**
   * The name of the file which contains the private key and the 
   * server certificate chain for <b>RSA</b> cipher suites.
   */
  public static final String 
  RSA_KEY_AND_CERTIFICATE_P = "iaik.jigsaw.ssl.rsa.keyAndCertificate";
  
  /**
   * The name of the file which contains the private key and the 
   * server certificate chain for <b>DSA</b> cipher suites.
   */
  public static final String 
  DSA_KEY_AND_CERTIFICATE_P = "iaik.jigsaw.ssl.dsa.keyAndCertificate";
  
  /**
   * The name of the file which contains the private key and the 
   * server certificate chain for <b>DH</b> cipher suites.
   */
  public static final String 
  DH_KEY_AND_CERTIFICATE_P = "iaik.jigsaw.ssl.dh.keyAndCertificate";
  
  /**
   * The name of the file containing the temporary RSA key pair or "generate"
   * if a new key pair shall be generated.
   */
  public static final String 
  TEMP_RSA_KEY_P = "iaik.jigsaw.ssl.tempRSAKey";
  
  /**
   * The string "generate" if new parameters shall be generated. If not
   * specified pre-generated values will be taken.
   */
  public static final String 
  DH_PARAMETERS_P = "iaik.jigsaw.ssl.dhPararmeters";
  
  /**
   * The cipher suites properties. Possible values are:
   * <nl>
   * <li>cipher.suite.ssl.rsa.with.null.md5
   * <li>cipher.suite.ssl.rsa.with.null.sha
   * <li>cipher.suite.ssl.rsa.export.with.rc4.40.md5
   * <li>cipher.suite.ssl.rsa.with.rc4.md5
   * <li>cipher.suite.ssl.rsa.with.rc4.sha
   * <li>cipher.suite.ssl.rsa.export.with.rc2.cbc.40.md5
   * <li>cipher.suite.ssl.rsa.with.idea.cbc.sha
   * <li>cipher.suite.ssl.rsa.export.with.des40.cbc.sha
   * <li>cipher.suite.ssl.rsa.with.des.cbc.sha
   * <li>cipher.suite.ssl.rsa.with.3des.ede.cbc.sha
   * <li>cipher.suite.ssl.dh.dss.export.with.des40.cbc.sha
   * <li>cipher.suite.ssl.dh.dss.with.des.cbc.sha
   * <li>cipher.suite.ssl.dh.dss.with.3des.ede.cbc.sha
   * <li>cipher.suite.ssl.dh.rsa.export.with.des40.cbc.sha
   * <li>cipher.suite.ssl.dh.rsa.with.des.cbc.sha
   * <li>cipher.suite.ssl.dh.rsa.with.3des.ede.cbc.sha
   * <li>cipher.suite.ssl.dhe.dss.export.with.des40.cbc.sha
   * <li>cipher.suite.ssl.dhe.dss.with.des.cbc.sha
   * <li>cipher.suite.ssl.dhe.dss.with.3des.ede.cbc.sha
   * <li>cipher.suite.ssl.dhe.rsa.export.with.des40.cbc.sha
   * <li>cipher.suite.ssl.dhe.rsa.with.des.cbc.sha
   * <li>cipher.suite.ssl.dhe.rsa.with.3des.ede.cbc.sha
   * <li>cipher.suite.ssl.dh.anon.export.with.rc4.40.md5
   * <li>cipher.suite.ssl.dh.anon.with.rc4.md5
   * <li>cipher.suite.ssl.dh.anon.export.with.des40.cbc.sha
   * <li>cipher.suite.ssl.dh.anon.with.des.cbc.sha
   * <li>cipher.suite.ssl.dh.anon.with.3des.ede.cbc.sha
   * </nl>
   */
  public static final Object[][] CIPHER_SUITES = {
    // 1. IDEA
    {"cipher.suite.ssl.rsa.with.idea.cbc.sha", CipherSuite.SSL_RSA_WITH_IDEA_CBC_SHA},

    // 2. Triple DES
    {"cipher.suite.ssl.rsa.with.3des.ede.cbc.sha", CipherSuite.SSL_RSA_WITH_3DES_EDE_CBC_SHA},
    {"cipher.suite.ssl.dh.dss.with.3des.ede.cbc.sha", CipherSuite.SSL_DH_DSS_WITH_3DES_EDE_CBC_SHA},
    {"cipher.suite.ssl.dh.rsa.with.3des.ede.cbc.sha", CipherSuite.SSL_DH_RSA_WITH_3DES_EDE_CBC_SHA},
    {"cipher.suite.ssl.dhe.dss.with.3des.ede.cbc.sha", CipherSuite.SSL_DHE_DSS_WITH_3DES_EDE_CBC_SHA},
    {"cipher.suite.ssl.dhe.rsa.with.3des.ede.cbc.sha", CipherSuite.SSL_DHE_RSA_WITH_3DES_EDE_CBC_SHA},

    // 3. RC4
    {"cipher.suite.ssl.rsa.with.rc4.md5", CipherSuite.SSL_RSA_WITH_RC4_MD5},
    {"cipher.suite.ssl.rsa.with.rc4.sha", CipherSuite.SSL_RSA_WITH_RC4_SHA},
    
    // 4. DES
    {"cipher.suite.ssl.rsa.with.des.cbc.sha", CipherSuite.SSL_RSA_WITH_DES_CBC_SHA},
    {"cipher.suite.ssl.dh.dss.with.des.cbc.sha", CipherSuite.SSL_DH_DSS_WITH_DES_CBC_SHA},
    {"cipher.suite.ssl.dh.rsa.with.des.cbc.sha", CipherSuite.SSL_DH_RSA_WITH_DES_CBC_SHA},
    {"cipher.suite.ssl.dhe.dss.with.des.cbc.sha", CipherSuite.SSL_DHE_DSS_WITH_DES_CBC_SHA},
    {"cipher.suite.ssl.dhe.rsa.with.des.cbc.sha", CipherSuite.SSL_DHE_RSA_WITH_DES_CBC_SHA},

    // 5. Export
    {"cipher.suite.ssl.rsa.export.with.rc4.40.md5", CipherSuite.SSL_RSA_EXPORT_WITH_RC4_40_MD5},
    {"cipher.suite.ssl.rsa.export.with.rc2.cbc.40.md5", CipherSuite.SSL_RSA_EXPORT_WITH_RC2_CBC_40_MD5},
    {"cipher.suite.ssl.rsa.export.with.des40.cbc.sha", CipherSuite.SSL_RSA_EXPORT_WITH_DES40_CBC_SHA},
    {"cipher.suite.ssl.dh.dss.export.with.des40.cbc.sha", CipherSuite.SSL_DH_DSS_EXPORT_WITH_DES40_CBC_SHA},
    {"cipher.suite.ssl.dh.rsa.export.with.des40.cbc.sha", CipherSuite.SSL_DH_RSA_EXPORT_WITH_DES40_CBC_SHA},
    {"cipher.suite.ssl.dhe.dss.export.with.des40.cbc.sha", CipherSuite.SSL_DHE_DSS_EXPORT_WITH_DES40_CBC_SHA},
    {"cipher.suite.ssl.dhe.rsa.export.with.des40.cbc.sha", CipherSuite.SSL_DHE_RSA_EXPORT_WITH_DES40_CBC_SHA},

    // 6. anon
    {"cipher.suite.ssl.dh.anon.with.rc4.md5", CipherSuite.SSL_DH_anon_WITH_RC4_MD5},
    {"cipher.suite.ssl.dh.anon.with.des.cbc.sha", CipherSuite.SSL_DH_anon_WITH_DES_CBC_SHA},
    {"cipher.suite.ssl.dh.anon.with.3des.ede.cbc.sha", CipherSuite.SSL_DH_anon_WITH_3DES_EDE_CBC_SHA},
    {"cipher.suite.ssl.dh.anon.export.with.rc4.40.md5", CipherSuite.SSL_DH_anon_EXPORT_WITH_RC4_40_MD5},
    {"cipher.suite.ssl.dh.anon.export.with.des40.cbc.sha", CipherSuite.SSL_DH_anon_EXPORT_WITH_DES40_CBC_SHA},
    
    // 7. no encryption
    {"cipher.suite.ssl.rsa.with.null.md5", CipherSuite.SSL_RSA_WITH_NULL_MD5},
    {"cipher.suite.ssl.rsa.with.null.sha", CipherSuite.SSL_RSA_WITH_NULL_SHA},
  };

  /**
   * Tries to create an InputStream form an specified file name.
   *
   * This method tries to open the file as absolute path or relative
   * to the server root.
   *
   * @param fileName the name of the file which shall be opened
   * @return a new FileInputStream if the file was found
   *
   * @exception FileNotFoundException
   *            if the file was not found
   */
  protected String getFileName(String fileName) throws FileNotFoundException {
    
    File file = new File(fileName);
    if (!file.exists())
      file = new File(ssl_server.getRootDirectory(), fileName);
      
    if (!file.exists())
      throw new FileNotFoundException(fileName);
      
    return file.getAbsolutePath();
  }

  /**
   * Some property have changed, update our setting.
   *
   * @param name the name of the property that has changed
   * @return true if we updated ourself successfully
   */
  public boolean propertyChanged(String name) {
    
    try {      
      
      if (name.equals(CLIENT_AUTHENTICATION_P)) {
        client_authentication = props.getBoolean(name, false);
        updateClientAuthentication();
      }
      else if (name.equals(CLIENT_REQUIRE_CERTIFICATE_P)) {
        require_certificate = props.getBoolean(name, false);
        trust_decider.setRequireCertificate(require_certificate);
      }
      else if (name.equals(CLIENT_REQUIRE_TRUSTED_ROOT_P)) {
        require_trusted_root = props.getBoolean(name, false);
        trust_decider.setRequireTrustedRoot(require_trusted_root);
      }
      else if (name.equals(TRUSTET_SIGNERS_P)) {
        trusted_signers_file_name = props.getString(name, trusted_signers_file_name);
        updateClientAuthentication();
      }
      else if (name.endsWith("keyAndCertificate")) {
        String file_name = props.getString(name, "certs/serverRSA.der");
        KeyAndCertificate kac = new KeyAndCertificate(getFileName(file_name));
        PrivateKeyInfo pki = (PrivateKeyInfo)kac.getPrivateKey();
      
        if (name.equals(RSA_KEY_AND_CERTIFICATE_P)) {
          ssl_context.setRSACertificate(kac.getCertificateChain(), (RSAPrivateKey)pki);
          ssl_server.errlog("RSA certificate set to: "+file_name);
          System.out.println("RSA certificate set to: "+file_name);
        }
        else if (name.equals(DSA_KEY_AND_CERTIFICATE_P)) {
          ssl_context.setDSACertificate(kac.getCertificateChain(), pki);
          ssl_server.errlog("DSA certificate set to: "+file_name);
          System.out.println("DSA certificate set to: "+file_name);
        }
        else if (name.equals(DH_KEY_AND_CERTIFICATE_P)) {
          ssl_context.setDHCertificate(kac.getCertificateChain(), pki);
          ssl_server.errlog("DH certificate set to: "+file_name);
          System.out.println("DH certificate set to: "+file_name);
        }
      }
      else if (name.equals(TEMP_RSA_KEY_P)) {
        String prop = props.getString(name, "");
        if (prop.equalsIgnoreCase("generate")) {
          System.out.println("generating new temporary RSA key...");
          
      		KeyPairGenerator generator = KeyPairGenerator.getInstance("RSA");
      		generator.initialize(512);
          ssl_context.setRSATempKeyPair(generator.generateKeyPair());
        }
        else {
          RSAPrivateKey pk = new RSAPrivateKey(new FileInputStream(prop));
          ssl_context.setRSATempKeyPair(new KeyPair(pk.getPublicKey(), pk));
        }
      }
      else if (name.equals(DH_PARAMETERS_P)) {
        String file_name = props.getString(name, "");
        DHParameterSpec dhparam = null;

        if (file_name.equalsIgnoreCase("generate")) {
          System.out.println("generating new DH parameters...");

          int primeLength = 512;      // length of prime p to generate
          int exponentLength = 504;   // length of random exponent (private)
          DHGenParameterSpec genParamSpec = new DHGenParameterSpec(primeLength, exponentLength); 

          AlgorithmParameterGenerator paramGenerator = AlgorithmParameterGenerator.getInstance("DH"); 
          paramGenerator.init(genParamSpec); 
          AlgorithmParameters parameters = paramGenerator.generateParameters(); 

          // we need the class for getParameterSpec(Class)
          Class classSpec = new DHParameterSpec(null, null).getClass();
          dhparam = (DHParameterSpec)parameters.getParameterSpec(classSpec); 
        }
        else {
          // pre-generated Diffi-Hellman parameters
          BigInteger p = new BigInteger("da583c16d9852289d0e4af756f4cca92dd4be533b804fb0fed94ef9c8a4403ed574650d36999db29d776276ba2d3d412e218f4dd1e084cf6d8003e7c4774e833", 16);
          BigInteger g = BigInteger.valueOf(2);
          dhparam = new DHParameterSpec(p, g);
        }
        
        // set the DH parameter for empherial and anon cipher suites
        ssl_context.setDHParameter(dhparam);
      }
      else if (name.startsWith("cipher.suite.ssl"))
        setupCipherSuites();
      else
        return super.propertyChanged(name);
        
    } catch (Exception ex) {
      ssl_server.errlog("error changing property: "+name+" because of: "+ex.toString());
      System.out.println("error changing property: "+name+" because of: "+ex.toString());
      return false;
    }
    
    return true;
  }

  /**
   * Initialize the properties for the first time when the server is started.
   */
  protected void initializeProperties() {

    if (!ssl_server.sslEnabled())
      return;
    
    System.out.println("creating new default SSLContext...");
    ssl_context = new SSLServerContext();
 
    // set the server trust decider
    trust_decider = new JigsawServerTrustDecider();
    ssl_context.setTrustDecider(trust_decider);

    // use the propertyChanged method to initialize the properties
  	if (props.getString(CLIENT_AUTHENTICATION_P, null) != null)
      propertyChanged(CLIENT_AUTHENTICATION_P);
  	if (props.getString(CLIENT_REQUIRE_CERTIFICATE_P, null) != null)
      propertyChanged(CLIENT_REQUIRE_CERTIFICATE_P);
  	if (props.getString(CLIENT_REQUIRE_TRUSTED_ROOT_P, null) != null)
      propertyChanged(CLIENT_REQUIRE_TRUSTED_ROOT_P);
  	if (props.getString(TRUSTET_SIGNERS_P, null) != null)
      propertyChanged(TRUSTET_SIGNERS_P);
  	if (props.getString(RSA_KEY_AND_CERTIFICATE_P, null) != null)
      propertyChanged(RSA_KEY_AND_CERTIFICATE_P);
  	if (props.getString(DSA_KEY_AND_CERTIFICATE_P, null) != null)
      propertyChanged(DSA_KEY_AND_CERTIFICATE_P);
  	if (props.getString(DH_KEY_AND_CERTIFICATE_P, null) != null)
      propertyChanged(DH_KEY_AND_CERTIFICATE_P);
  	if (props.getString(TEMP_RSA_KEY_P, null) != null)
      propertyChanged(TEMP_RSA_KEY_P);
  	if (props.getString(DH_PARAMETERS_P, null) != null)
      propertyChanged(DH_PARAMETERS_P);

    setupCipherSuites();
  }

  /**
   * Setup the cipher suites defined in the properties.
   */
  protected void setupCipherSuites() {

    Vector vector = new Vector();

    System.out.println("enable folowing cipher suites: ");
    for (int i=0; i<CIPHER_SUITES.length; i++) {
      if (props.getBoolean((String)CIPHER_SUITES[i][0], false)) {
        vector.addElement(CIPHER_SUITES[i][1]);
        System.out.println(((CipherSuite)CIPHER_SUITES[i][1]).getName());
      }
    }
    
    CipherSuite[] cs = new CipherSuite[vector.size()];
    vector.copyInto(cs);
    
    ssl_context.setEnabledCipherSuites(cs);
  }    

  /**
   * Updates the client authentication configuration if 
   * its properties change.
   */
  protected void updateClientAuthentication() {

    if (!client_authentication) {
      // no client authentication
      ssl_context.setRequireClientCertificate(null, null);
      return;
    }

    // if there is no file with the signer certificates we can not
    // generate the certificate request message
    if (trusted_signers_file_name == null) {
      return;
    }

    try {
      Vector names = new Vector();
      Vector certs = new Vector();
      
      FileInputStream is = new FileInputStream(trusted_signers_file_name);
      while (is.available() > 10) {
        X509Certificate cert = new X509Certificate(is);
        certs.addElement(cert);
        names.addElement(cert.getSubjectDN());
      }      

      X509Certificate[] cert_list = new X509Certificate[certs.size()];
      certs.copyInto(cert_list);
      trust_decider.setTrustedSigners(cert_list);
      
      // require a client certificate ?
      Name[] accepted_CAs = new Name[names.size()];
      names.copyInto(accepted_CAs);
      
      byte[] types = new byte[4];
      types[0] = ClientTrustDecider.rsa_sign;
      types[1] = ClientTrustDecider.dss_sign;
      types[2] = ClientTrustDecider.rsa_ephemeral_dh;
      types[3] = ClientTrustDecider.dss_ephemeral_dh;

      ssl_context.setRequireClientCertificate(types, accepted_CAs);
      System.out.println("Trusted signers set to: "+trusted_signers_file_name);
      
    } catch (FileNotFoundException ex) {
      ssl_server.errlog("Trusted signers file "+trusted_signers_file_name+" not found!");
      System.out.println("Trusted signers file "+trusted_signers_file_name+" not found!");
    } catch (IOException ex) {
      ssl_server.errlog("I/O error while reading file: "+trusted_signers_file_name+" ["+ex.toString()+"]");
      System.out.println("I/O error while reading file: "+trusted_signers_file_name+" ["+ex.toString()+"]");
    } catch (CertificateException ex) {
      ssl_server.errlog("Unable to parse certificate from file: "+trusted_signers_file_name+" ["+ex.toString()+"]");
      System.out.println("Unable to parse certificate from file: "+trusted_signers_file_name+" ["+ex.toString()+"]");
    }
  }

  /**
   * This is the only way to create s SSLSocketClient instead of a SocketClient.
   */
  protected synchronized SocketClientState addClient (boolean free) {
  	// Create a new client.
  	csList                   = new SocketClientState(csList);
  	SocketClientState cs     = csList;
  	SocketClient      client = new SSLSocketClient(server, this, cs);
  	cs.client = client;
  	clientCount++;
  	clientEstim++;
  	// Plug into free LRU if required:
  	if ( free ) {
	    cs.status = SocketClientState.C_FREE;
	    freeList.toHead(cs);
	    freeCount++;
  	}
  	return cs ;
  }

  /**
   * Creates the server socket.
   * If SSL is enabled a SSLServerSocket is created.
   */
  public ServerSocket createServerSocket() throws IOException {
    if (ssl_server.sslEnabled())
    	return new SSLServerSocket(server.getPort(), 128, ssl_context);
  	else
    	return super.createServerSocket();
  }
  
  /**
   * Initialize this client factory. If SSL is enabled a new
   * SSL configuration PropertySet is registered.
   */
  public void initialize(httpd server) {
    
    super.initialize(server);
    ssl_server = ((httpsd)server);
    initializeProperties();

    if (ssl_server.sslEnabled()) {
    	PropertySet set = new SSLPropertySet("SSLprops", ssl_server);
	    ssl_server.registerPropertySet(set);
	  }
  }
}